#!/usr/bin/python
## Copyright (C) 2003 Red Hat, Inc.
## Copyright (C) 2003 Dan Walsh <dwalsh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import gettext
from netboot_util import *
from pxeos import *

# I18N
gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
gettext.textdomain(PROGNAME)

try:
    gettext.install(PROGNAME, "/usr/share/locale", 1)
except IOError:
    import __builtin__
    __builtin__.__dict__['_'] = unicode

def hex2ip(hex):
    """ hex2ip will throw an exception on a badly formed ip"""
    ret=""
    for i in xrange((len(hex)+1)/2):
        hexno = '0x'+hex[2*i:2*i+2]
        if ret != "":
            ret = ret + "."
        ret = ret+ ("%d" % eval(hexno))
    return ret

def getPXEFiles():
    l={}
    if os.path.isdir(PXECFGDIR):
        for i in os.listdir(PXECFGDIR):
            try:
                l[i]=hex2ip(i)
            except:
                pass
    return l

class pxeboot:
    Args = [ 'initrd', 'root', 'init', 'ks', 'method', 'ip', 'console',
             'ramdisk_size', 'NFSROOT', 'SNAPSHOT', 'ETHERNET', 'NISDOMAIN' ]        
           # ^- kernel boot arguments that cannot be set by user           
    def __init__(self,val):
        try:
            s=socket.inet_aton(val)
            ip=val
        except socket.error,message:
            ip=socket.gethostbyname(val)
        self.ip=ip
        self.path = PXECFGDIR + self.getHEXIP().strip()
        self.setKickstart("")
        self.setSnapshot("")
        self.append=""
        self.ramdisk = None
        self.setEthernet("eth0")
        self.os=None
        self.setSerial(FALSE)
        self.setSerialUseSystemConsoleAlso(FALSE)
        self.setSerialSpeed(9600)
        self.setSerialParity('n')
        self.setSerialDataBits('8')
        self.NISdomain=None
        self.disableSnapshotGeneration=False
        self.remoteLogHost=None
        self.dict={}
        try:
            self.load()
        except:
            pass
        
    def __getNFSRoot(self,pxeos):
        return "root=/dev/ram0 init=disklessrc NFSROOT=%s:%s" % (pxeos.getServer(),pxeos.getLocation())

    def getIP(self):
        return self.ip;
    
    def getHostname(self, not_subnet=None):
        ipa=self.ip.split(".")
        ip=0
        octets=0
        if not_subnet == None:
            not_subnet = self.ip
        else:
            not_subnet = None
        try:
            for o in ipa:
                o = int(o)
                if (o < 0) or (o > 255):
                    raise ValueError
                octets += 1
                ip <<= 8
                ip |= o
        except ValueError:
            ip = 0
            octets = 0
        if (octets == 4) and ((ip & 0xff) != 0):
            # valid IP, not a subnet:
            try:
                ip = socket.gethostbyaddr(self.ip)[0]
                if len(ip):
                    not_subnet = ip
            except:
                pass
        if not_subnet == None:
            return self.ip
        else:
            return not_subnet
    
    def getEthernet(self):
        return self.ethernet

    def setEthernet(self,ethernet):
        self.ethernet=ethernet
    
    def __getEthernetString(self):
        return "ETHERNET="+self.getEthernet()

    def isSerial(self):
        return self.serial

    def setSerial(self,serial):
        self.serial=serial

    def setSerialUseSystemConsoleAlso(self,serial):
        self.serialUseSystemConsoleAlso=serial

    def getSerialUseSystemConsoleAlso(self):
        return self.serialUseSystemConsoleAlso

    def setSerialSpeed(self,speedString):
        bauds = [ 1200, 1800, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400 ]
        s = int(speedString)
        if s < len(bauds):
            s = bauds[ s ]
        else:            
            for i in range(0,len(bauds)):
                if (s <= bauds[i]) or ((i+1) == len(bauds)):
                    s = bauds[i]
                    break
            if i == len(bauds):
                s=bauds[4]
        self.serialSpeed = s

    def getSerialSpeed(self, type):
        if type == str:
            return str(self.serialSpeed)
        elif type==int:
            # Only used by GUI to get index of baud selection
            bauds = [ 1200, 1800, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400 ]
            for i in range(0,len(bauds)):
                if (self.serialSpeed == bauds[i]) or ((i+1) == len(bauds)):
                    return i
            self.serialSpeed = bauds[4]
            return 4

    def setSerialParity(self, v):
        if type(v) == str:
            if not v in ('n','e','o' ):
                v = 'n'            
            self.serialParity = v
        elif type(v) == int :
            pars=['n','e','o']
            if (v > 0) and (v < len(pars)):
                self.serialParity = pars[v]
            else:
                self.serialParity = 'n'
        else:
            self.serialParity = 'n'
    
    def getSerialParity(self, type):
        if type == str:
            return self.serialParity
        else:
            if self.serialParity == 'n':
                return 0
            elif self.serialParity == 'e':
                return 1
            elif self.serialParity == 'o':
                return 2
            else:
                return 0

    def setSerialDataBits(self,v):
        if type(v) == str:
            if not v in ('8','7','6','5'):
                v='8'
            self.serialDataBits = v
        elif type(v) == int:
            bits=['8','7','6','5']
            if ( v > 0 ) and ( v < len(bits) ):
                self.serialDataBits = bits[v]
            elif ( v > len(bits) ) and (v <= 8):
                self.serialDataBits = str(v)
            else:
                self.serialDataBits = '8'
        else:
            self.serialDataBits = '8'

    def getSerialDataBits(self, type):
        if type == str:
            return self.serialDataBits
        else:
            csMap={'8':0,'7':1,'6':2,'5':3}
            return csMap[self.serialDataBits]
    
    def __getConsoleString(self):
        if self.isSerial():
            if self.serialUseSystemConsoleAlso:
                return "console=ttyS0" + "," + self.getSerialSpeed(str) + self.getSerialParity(str) + self.getSerialDataBits(str) + " console=tty0"
            else:
                return "console=ttyS0" + "," + self.getSerialSpeed(str) + self.getSerialParity(str) + self.getSerialDataBits(str)

        else:
            return ""
    
    def setNISdomain(self, nisdomain):
        if (nisdomain == None) or (len(nisdomain)>0):
            self.NISdomain = nisdomain
        else:
            self.NISdomain = '(none)'

    def getNISdomain(self):
        return self.NISdomain

    def __getNISdomainString(self):
        if self.NISdomain != None:
            return 'NISDOMAIN=%s' % self.NISdomain
        else:
            return ''
    
    def setRamdisk(self,ramdisk):
        self.ramdisk=int(ramdisk)

    def getRamdisk(self, OS=None):
        if self.ramdisk == None:
            if OS == None:
                OS = self.getpxeos()
            if OS != None :
                self.ramdisk = OS.getRamDiskSize()
            else:
                self.ramdisk = 0
        return self.ramdisk

    def __getRamdiskString(self, OS=None):
        r = self.getRamdisk(OS) 
        if r == 0:
            return ""
        else:
            return "ramdisk_size="+str(r)

    def setKickstart(self,kickstart):
        self.kickstart=kickstart

    def getKickstart(self):
        return self.kickstart
    
    def __getKickstartString(self):
        ks=self.getKickstart()
        if len(ks):
            return "ks=%s" % ks
        else:
            return ""
        
    def setSnapshot(self,snapshot):
        self.snapshot=snapshot

    def getSnapshot(self):
        if self.snapshot=="":
            return self.getHostname(not_subnet=True)
        else:
            return self.snapshot
    
    def __getSnapshotString(self):
        snapshot = self.getSnapshot()
        if snapshot == None:
            return ""
        else:
            return "SNAPSHOT="+self.snapshot

    def setOS(self,OS):
        if OS==None:
            raise ValueError,_("You must specify an Operating System")
        l=pxeOSList().getOSList()
        for o in l:
            if o.getName() == OS:
                self.os=OS
                return
        raise ValueError,_("%s is not PXE Operating System") % OS
        
    def getOS(self):
        return self.os
        
    def getpxeos(self):
        l=pxeOSList().getOSList()
        for o in l:
            if o.getName() == self.os:
                return o
        raise ValueError,_("%s is not PXE Operating System") % self.os
        
    def getHEXIP(self):
        out=""
        for i in str.split(self.ip,"."):
            out=out + ("%02x" % int(i))
        return str.upper(out)
    
    def getFilePath(self):
        return self.path

    def isDiskless(self):
        return self.getpxeos().isDiskless()
        
    def __getLabel(self):
        OS=pxeos()
        OS.setName(self.os)
        return OS.getNameDir()

    def out(self):
        pxeos=self.getpxeos()
        label=pxeos.getNameDir()
        return ("default %s\n\nlabel %s\n    kernel %s/vmlinuz\n%s" % (label,label,label, self.__getAppend(pxeos)))

    def __getAppend(self,pxeos):
        appendLine=""
        if pxeos.isDiskless():
            appendLine= "    append %s initrd=%s/initrd.img %s %s %s %s %s" % (self.__getConsoleString(), self.__getLabel(), self.__getNFSRoot(pxeos),self.__getRamdiskString(pxeos),self.__getEthernetString(),self.__getSnapshotString(),self.__getNISdomainString())
        else:
            ks=self.__getKickstartString()
            if ks == '':
                ks = pxeos.getLoaderArgs(self.dict)
            appendLine = "    append initrd=%s/initrd.img %s %s %s %s" % (self.__getLabel(), self.__getConsoleString(),self.__getRamdiskString(pxeos), ks, self.__getNISdomainString() )
        for var in self.dict.keys():
            if var not in pxeboot.Args:
                if self.dict[var] != None:
                    appendLine = "%s %s=%s" % (appendLine,var,self.dict[var])
                else:
                    appendLine = "%s %s" % (appendLine,var)
        return appendLine
        
    def save(self):
        f = open(self.path,"w")
        f.write(self.out())
        f.close()
        pxeos= self.getpxeos()
        if pxeos.isDiskless():
            snapshot = self.getSnapshot()
            # Even if we aren't generating the snapshot, then ensure that the root
            # contains all the files and directories in diskless/files, else client can't mount them:
            getstatusoutput('files='+INSTALLDIR+'/diskless/files; root='+pxeos.getLocation()+'/root; while read f; do  if [ ! -e $root/$f ]; then if [[ "$f" == "*/" ]]; then mkdir -p $root/$f; else mkdir -p $root/${f%/*}; touch $root/$f; fi; fi; done < $files >/dev/null 2>&1')
            if (snapshot != None) and (len(snapshot)) and (not self.disableSnapshotGeneration):
                getstatusoutput("%s/diskless/syncfiles %s %s %s " % (INSTALLDIR, pxeos.getLocation(), self.getSnapshot(), pxeos.getKernel()))
            if (self.NISdomain != None) and ( len(self.NISdomain)>0 ) and ( self.NISdomain != '(none)' ):
                (e,o)=getstatusoutput(
                    "cd %s/root/etc/sysconfig || exit 1; if /bin/egrep -q '^NISDOMAIN' network; then t=`/bin/mktemp /tmp/XXXXXX`; grep -v '^NISDOMAIN' < network > $t; mv -f $t network; fi; echo 'NISDOMAIN=%s' >> network" % (pxeos.getLocation(), self.NISdomain))
            else :
                (e,o)=getstatusoutput(
                    "cd %s/root/etc/sysconfig || exit 1; if /bin/egrep -q '^NISDOMAIN' network; then t=`/bin/mktemp /tmp/XXXXXX`; grep -v '^NISDOMAIN' < network > $t; mv -f $t network; fi" % pxeos.getLocation() )
            
            (e,o)=getstatusoutput('/bin/egrep -q "^# syslog.conf maintained by system-config-netboot" ' +
                                  pxeos.getLocation() + '/root/etc/syslog.conf')
            if e == 0 :
                if self.remoteLogHost != None :                
                    if ( len(self.remoteLogHost) == 0 ):
                        self.remoteLogHost = pxeos.getServer()                
                    getstatusoutput(
                    'echo -e "# syslog.conf maintained by system-config-netboot\n# If you wish to maintain your own syslog.conf, remove the above line\n*.*	@' + self.remoteLogHost + '" > ' + pxeos.getLocation() + '/root/etc/syslog.conf' )
                else:
                    getstatusoutput(
                    'echo -e "# syslog.conf maintained by system-config-netboot\n# If you wish to maintain your own syslog.conf, remove the above line\n*.*	/var/log/messages" > ' + pxeos.getLocation() + '/root/etc/syslog.conf' )
    
    def handleAppend(self,append):        
        for i in append:
            # Extract all the fields we might set; users are not allowed
            # to set these boot arguments - all others go in dict:
            fields=i.split("=")            
            if fields[0]=="console":
                tsp=fields[1]
                tty=''
                speed=None
                parity=None
                dataBits=None
                if len(tsp) and (tsp[0:3] == 'tty'):
                    c=tsp.find(',')
                    if c > 3 :
                        tty=tsp[0:c]
                        parityIndex=-1
                        if tsp[-1] in ('8','7','6','5'):
                            dataBits=tsp[-1]
                            parityIndex=-2
                        if tsp[parityIndex] in ('n','e','o'):
                            parity=tsp[parityIndex]
                            tsp = tsp[0:parityIndex]                        
                        speed = tsp[c+1:]
                    else:
                        tty = tsp
                if tty[0:4]=="ttyS":
                    self.setSerial(1)
                    if speed != None:
                        self.setSerialSpeed(speed)
                    if parity != None:
                        self.setSerialParity(parity)
                    if dataBits != None:
                        self.setSerialDataBits(dataBits)
                elif tty=="tty0":
                    self.setSerialUseSystemConsoleAlso(1);
                continue
            elif fields[0] == 'NISDOMAIN':
                self.setNISdomain(fields[1])
            elif fields[0] in pxeboot.Args:
            # we do not allow these args to be set by user:
                continue
            # user specified boot argument:
            if len(fields) > 1:
                self.dict[fields[0]]=fields[1]
            else:
                self.dict[fields[0]]=None

    def load(self):
        f = open(self.path,"r")
        lines = f.readlines()
        for i in lines:
            fields=i.split()
            if len(fields) > 0:
                if fields[0]=="label":
                    # Don't disallow names containing "_" !
                    # k=fields[1].split("_")
                    # OS=""
                    # for s in k:
                    #    OS=OS+s+" "
                    # self.setOS(OS.strip())
                    self.setOS(fields[1].strip())
                if fields[0]=="append":
                    self.handleAppend(fields[1:])
        f.close()
        self.getpxeos()
        
    def delete(self):
        pxeos= self.getpxeos()
        os.system("rm -rf %s/snapshot/%s" % (pxeos.getLocation(), self.getIP()))
        os.remove(self.path)

def Usage():
    print _("pxeboot - Python pxeboot commandline tool\n\nUsage: pxeboot [-a|-m|-d|-l] [-h] [ -K <kickstart> ] [-S <snapshotname> ] [-r <ramdisk>] [-s] [-e <ethernet>] [-O <os>] hostname | IPAddress")

if __name__ == "__main__":
    import sys
    import signal
    import getopt
    cmdline = sys.argv[1:]
    sys.argv = sys.argv[:1]

    if os.getuid() != 0:
        print _("Please restart %s with root permissions!") % (sys.argv[0])
        sys.exit(10)
        
    signal.signal (signal.SIGINT, signal.SIG_DFL)
    class BadUsage: pass
    if len(cmdline)==0:
        Usage()
        sys.exit(10)
        
    ip = cmdline[-1]
    kickstart=None
    snapshot=None
    ramdisk=None
    ethernet="eth0"
    serial=FALSE
    OS=None
    add=FALSE
    modify=FALSE
    delete=FALSE
    progname = os.path.basename(sys.argv[0])

    try:
        opts, args = getopt.getopt(cmdline, "amdK:S:r:se:O:lh",
            ["add","modify","delete", "kickstart=","snapshot=","ramdisk=","serial", "ethernet=","os=","list","help"])
        for opt, val in opts:
            
            if opt == '-a' or opt == '--add':
                add=TRUE

            if opt == '-m' or opt == '--modify':
                modify=TRUE

            if opt == '-d' or opt == '--delete':
                delete=TRUE

            if opt == '-K' or opt == '--kickstart':
                kickstart=val

            if opt == '-S' or opt == '--snapshot':
                snapshot=val

            if opt == '-r' or opt == '--ramdisk':
                ramdisk=val

            if opt == '-e' or opt == '--ethernet':
                ethernet=val

            if opt == '-s' or opt == '--serial':
                serial=TRUE

            if opt == '-O' or opt == '--os':
                OS=val

            if opt == '-l' or opt == '--list':
                for i in getPXEFiles().values():
                    pxe=pxeboot(i)
                    try:
                        print pxe.out()
                    except ValueError,message:
                        print _("\nError:Invalid PXE Boot File %s%s\n%s\n" % (PXECFGDIR,i,message))
                        continue
                        
                sys.exit(0)

            if opt == '-h' or opt == '--help':
                Usage()
                sys.exit(0)

    except (getopt.error, BadUsage):
        Usage()
        sys.exit(1)

    if (add and delete) or (add and modify) or (modify and delete):
        Usage()
        sys.exit(1)

    try:
        if delete:
            pxe=pxeboot(ip)
            pxe.delete()
            sys.exit(0)

        if add or modify:
            if add:
                for i in getPXEFiles().values():
                    if i==ip:
                        print "%s already exists" % ip
                        sys.exit(0)
                    
                pxe=pxeboot(ip)
        
            if modify:
                found=FALSE
                for i in getPXEFiles().values():
                    if i==ip:
                        found=TRUE
                        break
                if not found:
                    print "%s does not exist" % ip
                    sys.exit(0)
                
            pxe=pxeboot(ip)
            if kickstart != None:
                pxe.setKickstart(kickstart)
            if snapshot != None:
                pxe.setSnapshot(snapshot)
            if ramdisk != None:
                pxe.setRamdisk(ramdisk)
            if ethernet != None:
                pxe.setEthernet(ethernet)
            pxe.setSerial(serial)
            pxe.setOS(OS)
            pxe.save()
            sys.exit(0)
            
        found=FALSE
        for i in getPXEFiles().values():
            if i==ip:
                found=TRUE
                break
        if not found:
            print "%s does not exist" % ip
            sys.exit(0)
        
        pxe=pxeboot(ip)
        pxe.out()
        
    except ValueError,e:
        print e
        sys.exit(1)
        
